// Project: ESP32-Doorbell          V2.0 (jan-2021)
// Programmers: Jos van der Zande
//              Paul Hermans
//
// Main ESP32-Doorbell module
//


#include <WiFi.h>
#include <SPIFFS.h>
#include "Main.h"
#include "WebServer.h"
#include <ArduinoOTA.h>
#include <ESPmDNS.h>
#include <HTTPClient.h>
#include "cam.h"
#include "FS.h"                // SD Card ESP32
#include "SD_MMC.h"            // SD Card ESP32
#include "soc/soc.h"           // Disable brownour problems
#include "soc/rtc_cntl_reg.h"  // Disable brownour problems
//#include "driver/rtc_io.h"
#include <esp_camera.h>
#include <OneWire.h>
#include "MillisTimer.h"
#include "cfg.h"
#include "hw.h"
#include "asc.h"
#include "session.h"

#ifdef use_telnet
TelnetSpy LOG;
#else 
#define Log Serial
#endif

#ifdef USE_FTP_SERVER
#include <FtpServer.h>
#endif

AscConnection *asyncc=NULL;

//#define BUTTOM_PUSH_STATE LOW

const char BUILD_MAIN[] = __DATE__ " " __TIME__;
//**************************************************************************************************************************************************
//**                                                           Setting Wifi credentials                                                           **
//**************************************************************************************************************************************************

char esp_board[20] = "none";			//ESP type selection
char esp_name[20] = "DoorBell32";		//Wifi SSID waarop ESP32-cam zich moet aanmelden.

//**************************************************************************************************************************************************
//**                                                                END SETTINGS END                                                              **
//**************************************************************************************************************************************************

uint8_t cardType = CARD_NONE;

AsyncWebServer webserver(80);
#ifdef USE_FTP_SERVER
FtpServer ftpSrv;
#endif

DNSServer dns;
ssdpAWS mySSDP(&webserver);
unsigned long startup_delay;
bool done_startup_delay=false;

/**************************************************************************
 *  Define One Wire with port
 * ************************************************************************/
OneWire ds(13);

// Define whether the pushbutton changes to LOW or HIGH when pushed
#if defined(BUTTOM_PUSH_STATE)
const int buttonPushedState = BUTTOM_PUSH_STATE;
#else
const int buttonPushedState = HIGH;
#endif

//unsigned long MQTT_lasttime;  // MQTT check lasttime
bool WifiOK = false;          // WiFi status
//bool mqtt_initdone = false;   // MQTT status
bool reboot = false;          // Pending reboot status
long rebootdelay = 0;         // used to calculate the delay

/**************************************************************************
 *  Storage and Communication vars for One Wire and Virtual ports
 * ************************************************************************/
// Define memory for one wire
uint8_t ow_addr[8];
bool ow_is_addr=false;
uint8_t ow_port_out=0;
uint8_t ow_pio_buf[13];  // Put everything in the buffer so we can compute CRC easily.
uint8_t ow_port=BIT_TAST|BIT_TAST_BOARD|BIT_SND_BUSY;
uint8_t ow_latch=0;
uint16_t ow_light=0;

/**************************************************************************
 *  Storage for last known Image on SD
 * ************************************************************************/
char LastImg[32];

MillisTimer DM_timer = MillisTimer();

/**************************************************************************
 *  Timer for One wire
 * ************************************************************************/
// Create a timer that fires every 100 milliseconds.
MillisTimer OW_timer = MillisTimer();

void PrintBytes(const uint8_t* addr, uint8_t count, bool newline=false) 
{
	for (uint8_t i = 0; i < count; i++)
	{
		Log.print(addr[i]>>4, HEX);
		Log.print(addr[i]&0x0f, HEX);
	}
	if (newline)
		Log.println();
}

void mount_sd()
{
		//Serial.println("Starting SD Card");
	if(!SD_MMC.begin("/sd",true)){
		Log.println("SD Card Mount Failed");
		cardType = CARD_NONE;
	}
	else
	{
		cardType = SD_MMC.cardType();
		if(cardType == CARD_NONE){
			Log.println("No SD Card attached");
		}
		else
		{
			Log.printf("SD Card attached %d\n\r",cardType);
		}
	}  

}

/**************************************************************************
 *  Take a picture and store on SD
 * ************************************************************************/

void MakePhoto(void)
{
	tm timeinfo;
	time_t now;
	char fileNumber[32];

	hw_do_flash(true);

	time(&now);
	localtime_r(&now, &timeinfo);

	camera_fb_t * frameBuffer = NULL;

	// Take Picture with Camera
	frameBuffer = esp_camera_fb_get();
	if (!frameBuffer) {
		Log.println("Camera capture failed");
	}
	else
	{
		bool do_write = false;
		sprintf(fileNumber, "/%04d-%02d-%02d-%02d-%02d-%02d.jpg",timeinfo.tm_year+1900,timeinfo.tm_mon+1,timeinfo.tm_mday, timeinfo.tm_hour,timeinfo.tm_min,timeinfo.tm_sec);

		fs::FS &fs = SD_MMC;
		File imgFile = fs.open(fileNumber, FILE_WRITE);
		if (!imgFile) {
			Log.println(F("Failed to open file in writing mode try to mount"));
			SD_MMC.end();
			cardType = CARD_NONE;
			mount_sd();
			if(cardType != CARD_NONE)
			{
				fs::FS &fs = SD_MMC;
				File imgFile = fs.open(fileNumber, FILE_WRITE);
				if (imgFile) {
					do_write=true;
				}
			}
		} 
		else 
		{
			do_write=true;
		}

		if(do_write)
		{
			imgFile.write(frameBuffer->buf, frameBuffer->len);
			Log.print("Saved ");
			Log.println(fileNumber);
			strcpy(LastImg,fileNumber+1);
			imgFile.close();
		}
		esp_camera_fb_return(frameBuffer);
	}

	hw_do_flash(false);
}

/**************************************************************************
 *  Buffer for single Char debug info and Output to Log im Full
 * ************************************************************************/
void w_ol(char x)
{
	static char ol[200];
	static uint8_t olp=0;

	ol[olp++]=x;
	if(olp>180)
	{
		ol[olp++]=0;
		Log.println(ol);
		olp=0;
	}
}

/**************************************************************************
 *  Search for One Wire devices
 * ************************************************************************/
bool OW_search(void)
{
	if(!ow_is_addr)
	{
		if (!ds.search(ow_addr)) 
		{
			Log.print("!");
			ds.reset_search();
			delay(20);
			return false;
		}

		if (OneWire::crc8(ow_addr, 7) != ow_addr[7]) 
		{
			Log.print("C");
			return false;
		}

		if (ow_addr[0] != 0x29) 
		{
			PrintBytes(ow_addr, 8);
			Log.print("N");
			return false;
		}

		ow_is_addr=true;
		return true;
	}
	return false;
}

/**************************************************************************
 *  Output to port defined on One Wire
 * ************************************************************************/
bool ow_write_port(uint8_t data)
{
	uint8_t buf[8];  // Put everything in the buffer.

	if(!ow_is_addr)
	{
		OW_search();
		// no address for one wire-device
		w_ol('n');
		return false;
	}

	// write the port to oneWire ; on error try 3 times
	buf[0] = 0x5A;    // write
	buf[1] = data;    // Value
	buf[2] = ~data;    // ~Value

	for(uint8_t i=0; i<3; i++)
	{
		ds.select(ow_addr);
		ds.write_bytes(buf, 3); // write this 3 bytes
		ds.read_bytes(buf+3, 2); // read two bytes first width AA second with the port-Value
		ds.reset();
		if(buf[3]==0xAA)
		{
			return true;
		}
		w_ol('f');
	}
	return false;
}

/**************************************************************************
 *  Clear the Latch for Falling or rising Portbits
 * ************************************************************************/
bool ow_clear_latch(void)
{
	uint8_t buf[8];  // Put everything in the buffer.

	// reset the latch ; on error try 3 times
	buf[0] = 0xC3;    // write

	for(uint8_t i=0; i<3; i++)
	{
		ds.select(ow_addr);
		ds.write_bytes(buf, 1); // write this 3 bytes
		ds.read_bytes(buf+1, 1); // read two bytes first width AA second with the port-Value
		ds.reset();
		if(buf[1]==0xAA)
		{
			ow_latch=0;
			return true;
		}
		w_ol('f');
	}
	return false;
}


/**************************************************************************
 *  Read the PIO and Light-Value
 * ************************************************************************/
bool ow_read_pio(void)
{
	if(!ow_is_addr)
	{
		OW_search();
		// no address for one wire-device
		w_ol('n');
		return false;
	}

	w_ol('.');

	ow_pio_buf[0] = 0xF0;    // Read PIO Registers
	ow_pio_buf[1] = 0x88;    // LSB address
	ow_pio_buf[2] = 0x00;    // MSB address

	for(uint8_t i=0; i<3; i++)
	{
		ds.select(ow_addr);
		ds.write_bytes(ow_pio_buf, 3);
		ds.read_bytes(ow_pio_buf+3, 10);     // 3 cmd bytes, 6 data bytes, 2 0xFF, 2 CRC16
		ds.reset();
		if (OneWire::check_crc16(ow_pio_buf, 11, &ow_pio_buf[11])) 
		{
			ow_port=ow_pio_buf[3];
			ow_latch=ow_pio_buf[5];
			ow_light=ow_pio_buf[9]+(ow_pio_buf[10]<<8);
//			PrintBytes(ow_pio_buf, 13,true);
			return true;
		}
//		PrintBytes(ow_pio_buf, 13,true);
		w_ol('c');
	}
//	PrintBytes(ow_pio_buf, 13,true);
	return false;
}

/**************************************************************************
 *  Timer Function for One Wire. Called every 100 ms
 * ************************************************************************/
void OW_tf(MillisTimer &mt)
{
	static uint8_t last_out=0x55;

	ow_read_pio();
	if(last_out!=ow_port_out)
	{
		if(ow_write_port(ow_port_out))
			last_out=ow_port_out;
	}
}

/**************************************************************************
 *  Timer Function Debug message
 * ************************************************************************/
void DM_tf(MillisTimer &mt)
{
	Log.printf("Light: %d\r\n",ow_light);
	Log.printf("Light Low: %d\r\n",cfg_global.light_dark_val);
	Log.printf("Light Heigh: %d\r\n",cfg_global.light_bright_val);
	Log.printf("Light Bool: %d\r\n",work_dark);
	Log.printf("Light Timer: %ld\r\n",light_timer);
}

/**************************************************************************
 *  Function called from loop to make the Logic
 * ************************************************************************/
void do_logic(void)
{
	static bool last_busy=true;
	static bool last_motion=false;
	static bool last_tast=true;

	OW_timer.run();

	if(seconds>light_timer)
	{
		// switch light off
		hw_do_light(false);
	}

	if(!IS_TAST && last_tast)
	{
		if(!hw_ring)
		{
			hw_ring=true;
			Ring_timer.setInterval(cfg_global.ring_to*1000);
			Ring_timer.reset();
			Ring_timer.start();
			// Door switch pressed
			if(cfg_global.switch_led_flash)
			{
				hw_led_flush();
			}
			if(cfg_global.switch_photo)
			{
				MakePhoto();
			}
			if(cfg_global.snd_on && cfg_global.snd_switch)
			{
				for(uint8_t i=0; i<CFG_SND_MAX;i++)
				{
					if(cfg_global.sound[i].ring)
					{
						hw_play_snd(atoi(cfg_global.sound[i].val.c_str()));
						break;
					}
				}
			}
			if(cfg_global.switch_call_on && cfg_global.switch_url !="")
			{
				String url=cfg_global.switch_url;
				url.replace("%img%", LastImg);

				asc_reset(asyncc);
				asc_del_headers(asyncc);
				asc_auth_basic(asyncc,cfg_global.switch_url_user.c_str(),cfg_global.switch_url_pw.c_str());
				asc_connect(asyncc,ASC_METHODE_GET,cfg_global.switch_url.c_str());
			}
		}
	}

	if(IS_MOTION && !last_motion)
	{
		if(!hw_motion)
		{
			hw_motion=true;
			Motion_timer.setInterval(cfg_global.motion_to*1000);
			Motion_timer.reset();
			Motion_timer.start();

			// switch light on?
			if((cfg_global.move_light && cfg_global.move_dark && work_dark) || (cfg_global.move_light && !cfg_global.move_dark))
			{
				// start or retrigger
				light_timer = seconds+cfg_global.move_delay;
				hw_do_light(true);
			}

			// bewegung erkannt
			if(cfg_global.move_photo)
			{
				// take foto ?
				MakePhoto();
			}
			if(cfg_global.move_call_on && cfg_global.move_url !="")
			{
				String url=cfg_global.move_url;
				url.replace("%img%", LastImg);

				asc_reset(asyncc);
				asc_del_headers(asyncc);
				asc_auth_basic(asyncc,cfg_global.move_url_user.c_str(),cfg_global.move_url_pw.c_str());
				asc_connect(asyncc,ASC_METHODE_GET,cfg_global.move_url.c_str());
			}
		}
		else
		{
			// retrigger timer
			Motion_timer.start();
		}
	}

	// detect if Busyline of sound is toggled
	if(IS_SOUND_BUSY && !last_busy)
	{
		if(work_do_open_after_snd)
		{
			// after sound open the door (if required)
			hw_open_door();
			work_do_open_after_snd=false;
		}
	}
	else if(!cfg_global.snd_on && work_do_open_after_snd)
	{
		// sound disabled, but button with open command is received
		hw_open_door();
		work_do_open_after_snd=false;
	}

	if(work_do_open)
	{
		// command 'open door' received
		if(!IS_OPEN_BUSY)
		{
			hw_open_door();
		}
		work_do_open=false;
	}

	if(ow_latch)
	{
		ow_clear_latch();
		Log.printf("Latch clear %02X\n\r",ow_port);
	}
	last_busy=IS_SOUND_BUSY;
	last_motion=IS_MOTION;
	last_tast=IS_TAST;
}

/**************************************************************************
 *  Arduino Setup
 * ************************************************************************/
void setup() {
	WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0); //disable brownout detector
	esp_log_level_set("*", ESP_LOG_NONE);
	esp_log_level_set("wifi", ESP_LOG_NONE);

	*LastImg='\0';
	//EEPROM.begin(200);
	Log.begin(9600);
	Log.setSerial(NULL);
//	Serial.begin(9600);

	// Log.setDebugOutput(true);

	// Initialize SPIFFS
	if (!SPIFFS.begin(true))
		Log.println(F("An Error has occurred while mounting SPIFFS"));

	// restore previous ESP saved settings
//	Restore_ESPConfig_from_SPIFFS();

	/**************************************************************************
	 *  Setup wifimanager. If valid connection parms this is transparent.
	 *  if no connection it starts a setup-Page for connection params
	 * ************************************************************************/
	//WiFiManager
	//Local intialization. Once its business is done, there is no need to keep it around
	AsyncWiFiManager wifiManager(&webserver, &dns);
	//reset saved settings
	//  wifiManager.resetSettings();
	// Previous line doesn't always work so this is another option to erase the EEPROM and all saved settings
	//  pio run --target erase

	// Try connecting to previous saved WiFI settings or else start as AP
	wifiManager.autoConnect(cfg_global.un.c_str(), cfg_global.up.c_str());
	//if you get here you have connected to the WiFi
	WifiOK = true;

	// Setup multi DNS (Bonjour)

	/**************************************************************************
	 *  setup MDNS
	 * ************************************************************************/
	if (MDNS.begin(esp_name)) 
	{
		Log.println("MDNS responder started");
		MDNS.addService("http", "tcp", 80);
		MDNS.addService("telnet", "tcp", 23);
		MDNS.addService("ota", "tcp", 8266);
	} 
	else 
	{
		Log.println("Error setting up MDNS responder!");
		delay(1000);
	}

#ifdef USE_OTA
	/**************************************************************************
	 *  Set OTA and Prepare the handling
	 * ************************************************************************/
	ArduinoOTA.setHostname(esp_name);

	ArduinoOTA.onStart([]() { Log.println("Start"); });
	ArduinoOTA.onEnd([]() { Log.println("\nEnd"); });
	ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) 
	{
		uint8_t p;
		p = (uint16_t)(progress / (total / 100));
		Log.printf("Progress: %u%%\r", p);
#ifdef use_telnet
		LOG.handle();
#endif

	});
	ArduinoOTA.onError([](ota_error_t error) 
	{
		Log.printf("Error[%u]: ", error);
		if (error == OTA_AUTH_ERROR)
			Log.println("Auth Failed");
		else if (error == OTA_BEGIN_ERROR)
			Log.println("Begin Failed");
		else if (error == OTA_CONNECT_ERROR)
			Log.println("Connect Failed");
		else if (error == OTA_RECEIVE_ERROR)
			Log.println("Receive Failed");
		else if (error == OTA_END_ERROR)
			Log.println("End Failed");
	});
	ArduinoOTA.begin();
#endif

	/**************************************************************************
	 *  Load the cfg-File and set the Vars
	 * ************************************************************************/
	cfg_LoadCfg();

	/**************************************************************************
	 *  set the NTP settings to get the right Time and Date
	 * ************************************************************************/
	// Get Network time
	//	const char *NTPpool = "192.168.168.99";
	//	const char *defaultTimezone = "CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";
	//	configTzTime(defaultTimezone, NTPpool);  //sets TZ and starts NTP sync

	configTzTime(cfg_global.sys_tz.c_str(), cfg_global.sys_ntp.c_str());  //sets TZ and starts NTP sync

	// wait max 5 secs till time is synced
	struct tm timeinfo;
	for (uint i = 0; i < 10; i++) {
		if (getLocalTime(&timeinfo,500)) {
			Log.println(F("Time synced.\n"));
			break;
		}
	}

	/**************************************************************************
	 *  Iit the Cam
	 * ************************************************************************/
	// Init Camera
	if(!initcamera())
	{
		// try once again
		initcamera();
	}

	/**************************************************************************
	 *  Mount the SD-Card
	 * ************************************************************************/
	mount_sd();

	// Set Button GPIO for INPUT and PULL UP/DOWN depending of the definition
	pinMode(BUTTON_GPIO_NUM, (buttonPushedState ? INPUT_PULLDOWN : INPUT_PULLUP));
	// Set LED output and optional the same flashes for the defined LED_BUILTIN
	pinMode(BUTTONLED_GPIO_NUM, OUTPUT);
	digitalWrite(BUTTONLED_GPIO_NUM, ON_LED_STATE);
	//Necessary for ESP_EYE (Jos:Not sure about this as this is for the other onboard buttons)
//	if (strcmp(esp_board, "ESP_EYE") == 0) {
//		pinMode(13, INPUT_PULLUP);
//		pinMode(14, INPUT_PULLUP);
//	}

	// start mqtt
//	if (!strcmp(SendProtocol, "mqtt")) {
//		Mqtt_begin();
//		mqtt_initdone = true;
//	}
//	MQTT_lasttime = millis();

	// Init WebServer
	WebServerInit(&webserver);

	//    .setAuthentication(esp_uname, esp_pass);


	/**************************************************************************
	 *  SSDP Service
	 * ************************************************************************/
	// Make ESP-CAM "known" in the network under it's ESP_NAME
	//initialize mDNS service
	//Define SSDP and model name
	const char *SSDP_Name = esp_name;
	const char *modelName = esp_board;
	const char *nVersion = BUILD_MAIN;
	const char *SerialNumber = ""; 
	const char *Manufacturer = "DoorBell32";
	const char *ManufacturerURL = "https://www.woodym.de";
	mySSDP.begin(SSDP_Name, SerialNumber, modelName, nVersion, Manufacturer, ManufacturerURL);
	Log.println(F("SSDP started\n"));

	/**************************************************************************
	 *  setup Pins 
	 * ************************************************************************/
	pinMode(4, OUTPUT);
	digitalWrite(4, LOW);
	pinMode(33, OUTPUT);
	digitalWrite(33, HIGH);


	/**************************************************************************
	 *  One Wire
	 * ************************************************************************/
	// init one wire
	OW_search();

	/**************************************************************************
	 *  Hardware and Virtual Ports from One Wire
	 * ************************************************************************/
	hw_init();

	/**************************************************************************
	 *  Sessions init
	 * ************************************************************************/
	session_init();

	/**************************************************************************
	 *  Timer set for One Wire. Only poll every 100ms
	 * ************************************************************************/

	OW_timer.setInterval(100);
	OW_timer.expiredHandler(OW_tf);
	OW_timer.start();

	/**************************************************************************
	 *  Timer set for Debug Message 20s
	 * ************************************************************************/

	DM_timer.setInterval(20001);
	DM_timer.expiredHandler(DM_tf);
	DM_timer.start();

	/**************************************************************************
	 *  Prepare Async connections
	 * ************************************************************************/
	asyncc=asc_new();

#ifdef USE_FTP_SERVER
	ftpSrv.begin(cfg_global.un.c_str(), cfg_global.up.c_str());   //username, password for ftp
#endif
	Log.print("Heap: ");
	Log.println(ESP.getFreeHeap());
	Log.print(" PsRam: ");
	Log.println(ESP.getFreePsram());

	// do stuff after 5 seconds at boot
	startup_delay=millis()+5000;
}

/**************************************************************************
 *  Arduino Loop
 * ************************************************************************/
void loop() {
	static uint8_t last_board_tast;
	static unsigned long time_board_tast;

	// update MDNS-Service
	// MDNS.update();

#ifdef USE_OTA
	/**************************************************************************
	 *  If OTA this will Handle the Request
	 * ************************************************************************/
	// OTA client code
	ArduinoOTA.handle();
#endif

#ifdef use_telnet
	LOG.handle();
#endif

	if(!done_startup_delay)
	{
		if(millis()>startup_delay)
		{
			// do things after boot delay like setting snd-volume
			done_startup_delay=true;
			hw_snd_vol(cfg_global.snd_vol);
			last_board_tast=IS_TAST_BOARD;
			time_board_tast=millis();
		}
	}
	else
	{
		// handle key on Attiny-Board
		if(!IS_TAST_BOARD && last_board_tast)
		{
			// rising edge for reset-tast
			time_board_tast=millis();
			Log.println(F("Board-Tast detected"));
		}
		else if(IS_TAST_BOARD && !last_board_tast)
		{
			// falling edge for reset-tast
			unsigned long tmp_time;

			Log.println(F("Board-Tast released"));

			tmp_time=millis()-time_board_tast;
			if(tmp_time >10000)
			{
				// erase wifi
				WiFi.disconnect(true,true);
				reboot = true;
				rebootdelay = millis();
			}
			else if(tmp_time >5000)
			{
				// open Door
				hw_open_door();
			}
			else if(tmp_time >1000)
			{
				// reboot
				reboot = true;
				rebootdelay = millis();
			}
			else
			{
				// play sound
				hw_play_snd(1);
			}
		}
		last_board_tast=IS_TAST_BOARD;
	}

	/**************************************************************************
	 *  If reboot is requestet, this handle the reboot
	 * ************************************************************************/
	// Don't do anything when shutting down and wait 1 second before rebooting
	if (reboot) {
		if (rebootdelay + 1000 > millis()) {
			Log.flush();
			SPIFFS.end();
			delay(500);
			ESP.restart();
			delay(2000);
		}
		return;
	}

	/**************************************************************************
 	*  do Logic for DoorBell
 	* ************************************************************************/
	do_logic();

	/**************************************************************************
	 *  do Logic for the hardware (like slower test for darkness)
	 * ************************************************************************/
	hw_loop();

	DM_timer.run();

#ifdef USE_FTP_SERVER
	ftpSrv.handleFTP();
#endif
}

